/* Copyright (C) 2002-2012 VA007219 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <conio.h>
#include <windows.h>
#include <tchar.h>
#include <crtdbg.h>
#include "std64.h"

#define STRINGIFY(s)   STRINGIFY_(s)
#define STRINGIFY_(s)  #s
#define LINE STRINGIFY(__LINE__)
#define TLINE _T(LINE)

#ifdef UNICODE
int iskanji(int x)
{
   if ( x >=0xd800 && x <= 0xdbff ) {
        return 1;
   } else {
       return 0;
   }
}
#undef _ftprintf
#define _ftprintf u_ftprintf
#undef _tprintf
#define _tprintf u_tprintf
int u_ftprintf(FILE *fp,const _TCHAR *format,...);
int u_tprintf(const _TCHAR *format,...);
#else
#define iskanji(x) IsDBCSLeadByte((BYTE)x)
#endif

int dbgprintf(_TCHAR *fmt,...);
int std_istalpha(int c);

static _TCHAR *version = _T("dirlist Version 6");
_TCHAR *pgname;

int lf;
int ef;
int pf;
int lc;

int namelen;
_TCHAR **extlist;
int extn;

#ifdef BCC32
_TCHAR *sterror2(int n)
{
static _TCHAR *r = NULL;
    _TCHAR *p;
    _TCHAR *q;
    int c;
int z;

    if ( r != NULL ) free(r);
    p = _tcserror(n);
    q = r = malloc((z=_tcslen(p)) * sizeof(_TCHAR));
    if ( q == NULL ) return p;
    while ( c = *p ) {
        if ( c == '\r' || c == '\n' ) {
            *q = 0;
            break;
        }
        *q = c;
        p++;
        q++;
    }
    return r;
}
#else
#define sterror2(x) _tcserror(x)
#endif

void help(void)
{
    _ftprintf(stderr,_T("usage: %s [dir]\n"),pgname);
    exit(1);
}

_TCHAR *nalloc(int n)
{
    if ( n < 256 ) {
        return malloc(256);
    } else {
        return malloc(n);
    }
}

_TCHAR *longpath(_TCHAR *file)
{
    _TCHAR *p;
    _TCHAR *q;
    _TCHAR *r;
    int c;
    int n;

    if ( file == NULL || *file == 0 ) {
        p = nalloc(2 * sizeof(_TCHAR));
        *p = 0;
        return p;
    }
    if ( *file == '\\' && *(file + 1) == '\\' && *(file + 2) == 0 ) { /* \\ */
        p = nalloc(3 * sizeof(_TCHAR));
        _stprintf(p,_T("\\\\"));
        return p;
    }
#ifdef UNICODE
    if ( *file == '\\' && *(file + 1) == '\\' && *(file + 2) == '?' && *(file + 3) == 0 ) { /* \\? */
        p = nalloc(4 * sizeof(_TCHAR));
        _stprintf(p,_T("\\\\?"));
        return p;
    }
    if ( *file == '\\' && *(file + 1) == '\\' && *(file + 2) == '?' && *(file + 3) == '\\' && *(file + 4) == 0 ) { /* \\?\ */
        p = nalloc(5 * sizeof(_TCHAR));
        _stprintf(p,_T("\\\\?\\"));
        return p;
    }
#endif
    p = nalloc((LONGMAX_PATH + 5) * sizeof(_TCHAR));
    if ( *file == '\\' && *(file + 1) == '\\' && *(file + 2) == '.' && *(file + 3) == '\\' ) { /* \\.\~ */
        lstrcpy(p,file);
        return p;
    }
    n = GetFullPathName(file,LONGMAX_PATH + 5,p,&r);
    if ( n == 0 ) {
        /* _ftprintf(errout,_T("%s: Invalid File Name. Get Last Error reports %d\n"),file,GetLastError ()); */
        lstrcpy(p,file);
        return p;
    }
#ifdef UNICODE
    if ( std_istalpha(*p) && *(p + 1) == ':' && *(p + 2) == '\\' ) { /* a:\~ -> \\?\A:\~ */
        c = *p;
        if ( _istlower(c) ) {
            *p = _totupper(c);
        }
        q = malloc((_tcslen(p) + 5) * sizeof(_TCHAR));
        _stprintf(q,_T("\\\\?\\%s"),p);
        return q;
    }
    if ( *p == '\\' && *(p + 1) == '\\' && *(p + 2) != '?' ) { /* \\pc \\pc\xxxx \\pc\xxxx\... -> \\?\~ */
        q = malloc((_tcslen(p) + 9) * sizeof(_TCHAR));
        _stprintf(q,_T("\\\\?\\UNC\\%s"),p + 2);
        return q;
    }
    if ( *p == '\\' && *(p + 1) == '\\' && *(p + 2) == '?' && *(p + 3) == '\\' && std_istalpha(*(p + 4)) && *(p + 5) == ':' && *(p + 6) == '\\' ) { /* \\?\A:\~ */
        c = *(p + 4);
        if ( _istlower(c) ) {
            *(p + 4) = _totupper(c);
        }
    }
#else
    if ( std_istalpha(*p) && *(p + 1) == ':' && *(p + 2) == '\\' ) { /* a:\~ -> \\?\A:\~ */
        c = *p;
        if ( _istlower(c) ) {
            *p = _totupper(c);
        }
    }
#endif
    return p;
}

void setdir(_TCHAR *buf,_TCHAR *path,_TCHAR *dir)
{
    if ( *path == 0 ) { /* "",dir -> "dir" */
        _stprintf(buf,_T("%s"),dir);
    } else if ( (lstrcmp(path,_T("\\")) == 0 || lstrcmp(path,_T("/")) == 0)) { /* "\",dir -> "\dir" */
        _stprintf(buf,_T("%s%s"),path,dir);
    } else if ( std_istalpha(*path) && *(path + 1) == ':' &&
        *(path + 2) == '\0') { /* "a:",dir  -> "a:dir" */
        _stprintf(buf,_T("%s%s"),path,dir);
    } else if ( std_istalpha(*path) && *(path + 1) == ':' &&
        (*(path + 2) == '\\' || *(path + 2) == '/') && *(path + 3) == '\0') { /* "a:\",dir  -> "a:\dir" */
        _stprintf(buf,_T("%s%s"),path,dir);
    } else {
        _stprintf(buf,_T("%s\\%s"),path,dir);
    }
}

_TCHAR *extname(_TCHAR *name)
{
    _TCHAR *p = name;
    _TCHAR *q;
    int     c;

    q = NULL;
    if ( std_istalpha(*p) && *(p + 1) == ':' ) p += 2;
    while ( (c = *p) != '\0' ) {
        if ( c == '.' ) {
            q = p;
        } else if ( iskanji(c) ) {
            if ( *(p + 1) != '\0' ) {
               p++;
            }
        } else if ( (c == '\\' || c == '/') ) {
            q = NULL;
        }
        p++;
    } 
    return q;
}

_TCHAR *nobs(_TCHAR *name)
{
    _TCHAR *p = name;
    _TCHAR *q;
    int c;

    while ( (c = *p) != '\0' ) {
        q = NULL;
        if ( c == '\\' ) {
            q = p;
        } else if ( iskanji(c) ) {
            if ( *(p + 1) != '\0' ) {
               p++;
            }
        }
        p++;
    } 
    if ( q != NULL ) {
        *q = 0;
// _ftprintf(stderr,TLINE _T("nobs=%s\n"),name);
    }
    return name;
}

_TCHAR *dirname(_TCHAR *p)
{
    _TCHAR *s;
    _TCHAR *q;
    int c;

    if ( (p = s = _tcsdup(p)) == NULL ) {
        _ftprintf(stderr,_T("%s: inadequate memory.\n"),pgname);
        exit(1);
    }
    if ( (*p == '\\' || *p == '/') && (*(p + 1) == '\\' || *(p + 1) == '/') ) { /* \\xxx */
        p += 2;
    } else if ( (*p == '\\' || *p == '/') ) { /* \xxx */
        p += 1;
    } else if ( _istalpha(*p) && *(p + 1) == ':' && (*(p + 2) == '\\' || *(p + 2) == '/') ) { /* a:\xxx */
        p += 3;
    } else if ( _istalpha(*p) && *(p + 1) == ':' ) { /* a:xxx */
        p += 2;
    }
    q = p;
    while ( (c = *p) != '\0' ) {
       if ( (c == '\\' || c == '/') ) {
           q = p;
       } else if ( iskanji(c) ) {
           if ( *(p + 1) != '\0' ) {
               p++;
           /*
           } else {
               _ftprintf(errout,_T("%s: %s: bad filename.\n"),pgname,s);
               exit (1);
           */
           }
       }
       p++;
    } 
    *q = '\0';
    return s;
}

void sublist(_TCHAR *srcdir)
{
    struct std_find_data64 buf;
    _TCHAR *p;
    long h;
    _TCHAR *src;
    struct std_stat64 sbuf;
    struct tm *t;
    int n;
    _TCHAR *ep;
    int exef;
int i;
int z;

    if ( *srcdir != 0 ) {

        struct std_find_data64 findbuf;
        int h;

        h = std_findfirst64(srcdir,&findbuf);
        if ( h == -1 ) {
            _ftprintf(stderr,_T("\"%s\" %s\n"),srcdir,sterror2(errno));
            return;
        } else {
            std_findclose(h);
        }
        if ( findbuf.w32_finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
            ;
        } else {
            _tprintf(_T("\"%s\", %I64d\n"),findbuf.name,findbuf.size);
            return;
        }
    }
    if ( *srcdir == 0 ) { /* "" -> "*.*" */
        p = malloc((z=(4)) * sizeof(_TCHAR));
        _stprintf_s(p,z,_T("*.*"));
    } else if ( (lstrcmp(srcdir,_T("\\")) == 0 || lstrcmp(srcdir,_T("/")) == 0) ) { /* "\" -> "\*.*" */
        p = malloc((z=(5)) * sizeof(_TCHAR));
        _stprintf_s(p,z,_T("\\*.*"));
    } else if ( std_istalpha(*srcdir) && *(srcdir + 1) == ':' && *(srcdir + 2) == '\0') { /* "a:"  -> "a:*.*" */
        p = malloc((z=(6)) * sizeof(_TCHAR));
        _stprintf_s(p,z,_T("%s*.*"),srcdir);
    } else if ( std_istalpha(*srcdir) && *(srcdir + 1) == ':' && (*(srcdir + 2) == '\\' || *(srcdir + 2) == '/') && *(srcdir + 3) == '\0') { /* "a:\"  -> "a:\*.*" */
        p = malloc((z=(7)) * sizeof(_TCHAR));
        _stprintf_s(p,z,_T("%s*.*"),srcdir);
    } else {
        p = malloc((z=(lstrlen(srcdir) + 5)) * sizeof(_TCHAR));
        _stprintf_s(p,z,_T("%s\\*.*"),srcdir);
    }
    if ( (h = std_findfirst64(p,&buf)) == -1 ) {
        _ftprintf(stderr,_T("%s: %s\n"),p,sterror2(errno));
    } else {
        do {
            if ( buf.attrib & _A_SUBDIR ) {
                if ( lstrcmp(buf.name,_T(".")) == 0 || lstrcmp(buf.name,_T("..")) == 0 ) continue;
                src = malloc((z=(lstrlen(srcdir) + lstrlen(buf.name) + 2)) * sizeof(_TCHAR));
                setdir(src,srcdir,buf.name);
                _tprintf(_T("\"%s\", dir\n"),src + namelen);
                sublist(src);
                free(src);
            } else {
                if ( ef ) {
                    if ( (ep = extname(buf.name)) == NULL ) ep = _T(".");
                    ep++;
                    exef = 0;
                    for ( n = 0 ; n < extn ; n ++ ) {
                        if ( lstrcmpi(ep,extlist[n]) == 0 ) {
                            exef = 1;
                            break;
                        }
                    }
                    if ( exef == 0 ) continue;
                }
                src = malloc((z=(lstrlen(srcdir) + lstrlen(buf.name) + 2)) * sizeof(_TCHAR));
                setdir(src,srcdir,buf.name);
                _tprintf(_T("\"%s\", %I64d\n"),src + namelen,buf.size);
                free(src);
            }
        } while ( std_findnext64(h,&buf) == 0 );
    }
    std_findclose(h);
    free(p);
}

void list(_TCHAR *arg)
{
    _TCHAR *file;
    _TCHAR *p = NULL;
    _TCHAR *q;
    _TCHAR *r;
    long h;
    int n;
    _TCHAR *ep;
    int exef;
    struct std_find_data64 findbuf;
    _TCHAR *wc;
int z;

    file = arg;
    p = nalloc((LONGMAX_PATH + 5) * sizeof(_TCHAR));
    n = GetFullPathName(file,LONGMAX_PATH + 5,p,&r);
    h = std_findfirst64(file,&findbuf);
// _ftprintf(stderr,TLINE _T("h=%d file=%s\n"),h,file);
    if ( h == -1 ) {
        if ( !( std_istalpha(*file) && *(file + 1) == ':' && (*(file + 2) == '\\' ) && *(file + 3) == '\0' ) &&
             !( *file == '\\' && *(file + 1) == '\\' && *(file + 2) == '?' && *(file + 3) == '\\' && std_istalpha(*(file+4)) && *(file+4 + 1) == ':' && *(file+4 + 2) == '\\' && *(file+4 + 3) == '\0' ) ) {
            file = nobs(p);
            h = std_findfirst64(file,&findbuf);
// _ftprintf(stderr,TLINE _T("h=%d file=%s\n"),h,file);
        }
    }
    if ( h == -1 ) {
        file = arg;
// _ftprintf(stderr,TLINE _T("h=%d file=%s\n"),h,file);
        if ( r == NULL ) {
            wc = _T("*.*");
        } else {
            wc = _T("\\*.*");
        }
        if ( *file == 0 ) { /* "" -> "*.*" */
            _stprintf_s(p,LONGMAX_PATH + 5,_T("*.*"));
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
        } else if ( std_istalpha(*file) && *(file + 1) == ':' && *(file + 2) == '\0' ) { /* "a:"  -> "a:*.*" */
            _stprintf_s(p,LONGMAX_PATH + 5,_T("%s*.*"),file);
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
        } else if ( std_istalpha(*file) && *(file + 1) == ':' && (*(file + 2) == '\\' ) && *(file + 3) == '\0' ) { /* "a:\"  -> "a:\*.*" */
            _stprintf_s(p,LONGMAX_PATH + 5,_T("%s*.*"),file);
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
        } else if ( *file == '\\' && *(file + 1) == '\\' && *(file + 2) != '?' && *(file + 3) != 0 ) { /* "\\~~"  -> "\\~~\*.*" */
            _stprintf_s(p,LONGMAX_PATH + 5,_T("%s%s"),file,wc);
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
        } else if ( *file == '\\' && *(file + 1) == '\\' && *(file + 2) == '?' && *(file + 3) == '\\' && *(file + 4) != 0 ) { /* \\?\~ */
// _ftprintf(stderr,TLINE _T("file=%s\n"),file);
            if ( std_istalpha(*(file+4)) && *(file+4 + 1) == ':' && *(file+4 + 2) == '\\' && *(file+4 + 3) == '\0' ) { /* "\\?\a:\"  -> "\\?\a:\*.*" */
                _stprintf_s(p,LONGMAX_PATH + 5,_T("%s*.*"),file);
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
            } else if ( _totlower(*(file + 4)) == 'u' && _totlower(*(file + 5)) == 'n' && _totlower(*(file + 6)) == 'c' && *(file + 7) == '\\' ) {  /* "\\?\unc\~"  -> "\\?\UNC\~\*.*" */
                _stprintf_s(p,LONGMAX_PATH + 5,_T("\\\\?\\UNC\\%s%s"),file + 8,wc);
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
            } else {
            _stprintf_s(p,LONGMAX_PATH + 5,_T("%s"),file);
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
            }            
        } else {
            _stprintf_s(p,LONGMAX_PATH + 5,_T("%s"),file);
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
        }
// _ftprintf(stderr,TLINE _T("p=%s\n"),p);
        file = p;
// _ftprintf(stderr,TLINE _T("wc=%s p=%s file=%s\n"),wc,p,file);
        h = std_findfirst64(file,&findbuf);
        if ( h == -1 ) {
            _ftprintf(stderr,_T("%s: %s\n"),arg,sterror2(errno));
            return;
        }
    }
    do {
        if ( findbuf.attrib & _A_SUBDIR ) {
            if ( lstrcmp(findbuf.name,_T(".")) == 0 || lstrcmp(findbuf.name,_T("..")) == 0 ) continue;
            p = malloc((z=(lstrlen(file) + lstrlen(findbuf.name) + 2)) * sizeof(_TCHAR));
            q = dirname(file);
            setdir(p,q,findbuf.name);
            free(q);
            namelen = lstrlen(p) + 1;
            sublist(p);
            free(p);
            p = NULL;
        } else {
            if ( ef ) {
                if ( (ep = extname(findbuf.name)) == NULL ) ep = _T(".");
                ep++;
                exef = 0;
                for ( n = 0 ; n < extn ; n ++ ) {
                    if ( lstrcmpi(ep,extlist[n]) == 0 ) {
                        exef = 1;
                        break;
                    }
                }
                if ( exef == 0 ) continue;
            }
            _tprintf(_T("\"%s\", %I64d\n"),findbuf.name,findbuf.size);
        }
    } while ( std_findnext64(h,&findbuf) == 0 );
    std_findclose(h);
    if ( p != NULL ) {
        free(p);
        p = NULL;
    }
}

int _tmain(int argc,_TCHAR **argv)
{
    int i;
    int c;
    _TCHAR *p;
    _TCHAR *path;

    _ftprintf(stderr,_T("%s Copyright (C) 2002-2012 VA007219\n"),version);
    pgname = argv[0];

    while ( argc >= 2 && (*argv[1] == '-' || *argv[1] == '/') ) { /* option */
	if ( lstrcmp(argv[1],_T("--")) == 0 ) { /* end of options */
	    ++argv ; --argc ;
	    break;
        } else if ( (c = *(p = argv[1])) == '/' || c == '-' ) {
            p++;
            while ( (c = _totupper(*(p++))) != 0 ) {
                if ( c == 'L' ) {
                    lf = 1;
                } else if ( c == 'P' ) {
                    lc = 0;
                    pf = 1;
                } else if ( c == '/' ) {
                    ;
                } else if ( c == '-' ) {
                    ;
                } else if ( c == 'H' ) {
                    help();
                } else if ( c == '?' ) {
                    help();
                } else {
                    _ftprintf(stderr,_T("%s: bad option %c\n"),pgname,c);
                    _ftprintf(stderr,_T("for more infomation: %s /?\n"),pgname);
                    exit(1);
                }
            }
        } else {
            _ftprintf(stderr,_T("%s: bad option %c\n"),pgname,c);
            _ftprintf(stderr,_T("for more infomation: %s /?\n"),pgname);
            exit(1);
        }
        ++argv ; --argc ;
    }

    if ( argc < 2 ) {
        namelen = 0;
        list(_T("*.*"));
    } else {
        for ( i = 1 ; i < argc ; ++i ) {
            if ( *argv[i] == 0 ) {
                /* "" */
                list(_T("*.*"));
            } else {
                path = longpath(argv[i]);
                list(path);
                free(path);
            }
        }
    }
    return 0;
}
